/*******************************************************************************
 *
 * libbasic 
 *
 * libbasic is a simple framework of foundation classes written in C++.
 *
 * The full text of the GNU General Public License can be found in the doc 
 * directory.
 *
 * Copyright (C) 2002, 2003 Stefan Hoefer
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact Stefan Hoefer via email at stefan@hoefer.ch
 *
 ******************************************************************************/

#ifndef BASICLIST_H

#define BASICLIST_H

#include "simplelist.h"
#include "unique_access.h"

template <class node>
class iterator;
template <class node>
class sorted_iterator;

/*! \class basic_list
 * \brief A basic container class (list) with a timestamp
 *
 * This is a very simple list class. It provides simple member manipulation commands (like adding and 
 * removing members to and from the list) and also keeps track of a time stamp. The only function 
 * manipulating the member array is \b flip.
 *  
 */

template <class node>
class basic_list : public CSimpleList<node>, public unique_access 
{
    friend class iterator<node>;
    friend class sorted_iterator<node>;
 protected:
/*! 
  \fn virtual int internal_equality(node x, node y)
  \brief A comparison function for internal purposes.
  \param x the first element to compare
  \param y the second element to compare
  \return 1 iff x and y are found to be equal, 0 else
 
  This function compares the two elements and tells whether they are equal or not. Classes inheriting
  basic_list may redefine this virtual implementation to give it special behaviour. The default 
  implementation returns 1, iff \e x == \e y
 */
    virtual int internal_equality(node x, node y) const;
 public:

    basic_list(unsigned int maxnumber = 10);
  
    ~basic_list();

/*!
  \fn void operator+=(node x)
  \brief Adds an element to the list.
  \param x The element to be added.  
 */
    void operator+=(node x);
/*!
  \fn int operator-=(node x)
  \brief Removes an element from the list.
  \param x The element to be removed from the list.
  \return 1 if an element was removed, 0 else.

  This function uses the \a internal_equality method in order to decide whether an element in the list is the same as \a x.
 */
    int operator-=(node x);
    void reset();

/*!
  \fn node in(node x)
  \brief Tests whether a given element is in the list or not.
  \param x The element to be tested for.
  \return A pointer to the element in the list, or NULL, if none was found.

  This function uses the \a internal equality method in order to decide whether an element is the same as \a x.
 */
    node in(node x) const;
  	
    void flip();
    void get_members(basic_list<node> const *l);
};

//---------------------------------------------------------------------------------------
// Implementation Area
//---------------------------------------------------------------------------------------

template <class node>
int basic_list<node>::internal_equality(node x, node y) const
{
    if(x==y)
    {
	return 1;
    }
    else
    {
	return 0;
    }
}

template <class node>
basic_list<node>::basic_list(unsigned int maxnumber) : CSimpleList<node>(maxnumber), unique_access()
{

}

template <class node>
basic_list<node>::~basic_list()
{

}

template <class node>
void basic_list<node>::operator+=(node x)
{
    CSimpleList<node>::operator+=(x);
    touch();
}

template <class node>
int basic_list<node>::operator-=(node x)
{
    touch();

    node *zeiger,*source,*dest;
    int i;

    // STH sandalle SF zeiger = get_field();
    zeiger = this->get_field();
    
    i = 0;
    // STH sandalle SF while(i<card()&&!internal_equality(*zeiger,x))
    while(i<this->card()&&!internal_equality(*zeiger,x)) // STH sandalle SF
    {
      	zeiger++;
      	i++;
    }

    //STH sandalle SF if(i<card())
    if(i<this->card())
    {
      	dest = zeiger;
      	source = zeiger+1;
      	i++;

      	while(i<card())
      	{
	    i++;
	    *dest++ = *source++;
      	}

      	// STH sandalle SF ReduceCardBy1();
      	this->ReduceCardBy1();

      	return 1;
    }
    else
    {
      	return 0;
    }
}

template <class node>
void basic_list<node>::reset()
{
    CSimpleList<node>::reset();
    touch();
}

template <class node>
node basic_list<node>::in(node x) const
{
    node *zeiger;
    int i;

    // STH sandalle SF zeiger = get_field();
    zeiger = this->get_field();
    i = 0;
      
    // STH sandalle SF while(i<card())
    while(i<this->card())
    {
	if(internal_equality(*zeiger,x))
	{
	    return(*zeiger);
	}
	i++;
	zeiger++;
    }
      
    return (node)0;
}

template <class node>
void basic_list<node>::flip()
{
    CSimpleList<node>::flip();
    touch();
}

template <class node>
void basic_list<node>::get_members(basic_list<node> const *l)
{
    CSimpleList<node>::get_members(l);
    touch();
}

#endif
