/*******************************************************************************
 *
 * libbasic
 *
 * libbasic is a simple framework of foundation classes written in C++.
 *
 * The full text of the GNU General Public License can be found in the doc 
 * directory.
 *
 * Copyright (C) 2002, 2003 Stefan Hoefer
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact Stefan Hoefer via email at stefan@hoefer.ch
 *
 ******************************************************************************/

#ifndef TREE_H

#define TREE_H

#include "simplelist.h"
#include "simpleiterator.h"
#include "identificator.h"

template <class T>
class CTreeNode : CSimpleList<CTreeNode<T> *>
{
 private:
    T m_Node;
    CTreeNode<T> *m_Parent;
 private:
    void CommonConstruct(CTreeNode<T> *pParent);
 protected:
    inline T *NodeAdress() { return &m_Node; }
    CTreeNode(IDENT id, CTreeNode<T> *pParent=NULL);
 public:
    inline T &Node() { return m_Node; }
    inline CTreeNode<T> *Parent() { return m_Parent;}
		
    // STH sandalle SF inline CTreeNode<T> *Child(int nIndex) { return operator[](nIndex); }
    inline CTreeNode<T> *Child(int nIndex) { return this->operator[](nIndex); }
		
    CTreeNode(CTreeNode<T> *Parent = NULL);
    virtual ~CTreeNode();
		
    virtual void operator+=(CTreeNode<T> *pTree);
    virtual void operator-=(CTreeNode<T> *pTree);
    bool IsChildOf(CTreeNode<T> *pTree);
    CTreeNode<T> *Root();

    virtual CTreeNode<T> *Copy(CTreeNode<T> *pNode);
    virtual bool Adopt(CTreeNode<T> *pNode);
    virtual bool IsDescendantOf(CTreeNode<T> *pNode);

    inline CSimpleList<CTreeNode<T> *> &Children() { return *this; } 
};

template <class T>
CTreeNode<T> *CTreeNode<T>::Copy(CTreeNode<T> *pNode)
{
    CTreeNode<T> *pNewNode = new CTreeNode<T>(*pNode);

    *this += pNewNode;

    int nIndex = 0;

    while (CTreeNode<T> *pChild = pNode->Child(nIndex++))
    {
	VERIFY(pNewNode->Copy(pChild));
    }

    return pNewNode;
}

template <class T>
bool CTreeNode<T>::Adopt(CTreeNode<T> *pNode)
{
    //ASSERT(Root() == pNode->Root());
    ASSERT(pNode->Parent() != NULL);

    if (!IsDescendantOf(pNode))
    {
	CTreeNode<T> *pParent = pNode->Parent();
	*pParent -= pNode;
	*this += pNode;
    }
    else
    {
	return false;
    }
}

template <class T>
bool CTreeNode<T>::IsDescendantOf(CTreeNode<T> *pNode)
{
    CTreeNode<T> *pParent = Parent();

    do
    {
	if (pParent == pNode)
	{
	    return true;
	}
    } while (pParent = pParent->Parent());

    return false;
}

template <class T>
CTreeNode<T> *CTreeNode<T>::Root()
{
    if (Parent() == NULL)
    {
	return this;
    }
    else
    {
	return Parent()->Root();
    }
}

template <class T>
bool CTreeNode<T>::IsChildOf(CTreeNode<T> *pTree)
{
    int nIndex = 0;

    while (CTreeNode<T> *pSubTree = pTree->Child(nIndex++))
    {
	if (pSubTree == this)
	{
	    return true;
	}
    }

    return false;
}

template <class T>
void CTreeNode<T>::operator-=(CTreeNode<T> *pTree)
{
    ASSERT(pTree->m_Parent == this);
    
    pTree->m_Parent = NULL;
    CSimpleList<CTreeNode<T> *>::operator-=(pTree);
}

template <class T>
void CTreeNode<T>::operator+=(CTreeNode<T> *pTree)
{
    ASSERT(pTree->m_Parent == NULL);

    pTree->m_Parent = this;

    CSimpleList<CTreeNode<T> *>::operator += (pTree);
}

template <class T>
CTreeNode<T>::CTreeNode(IDENT id, CTreeNode<T> *pParent)
: m_Node(id)
{
    CommonConstruct(pParent);
}

template <class T>
void CTreeNode<T>::CommonConstruct(CTreeNode<T> *Parent)
{
    m_Parent = NULL;

    if ( Parent != NULL )
    {
	*Parent += this;
    }

    ASSERT(m_Parent == Parent);	
}

template <class T>
CTreeNode<T>::CTreeNode(CTreeNode<T> *Parent)
{
    CommonConstruct(Parent);
}

template <class T>
CTreeNode<T>::~CTreeNode()
{
    if ( m_Parent )
    {
	*m_Parent -= this;
    }
	
    CTreeNode<T> *pTree;
    CSimpleIterator<CTreeNode<T> *> TreeIt;
	
    TreeIt << this;
	
    while ( pTree = TreeIt++ )
    {
	pTree->m_Parent = NULL;
	delete pTree;
    }
}

#endif









