/*******************************************************************************
 *
 * libbasic Version 0.0.1
 *
 * libbasic is a simple framework of foundation classes written in C++.
 *
 * The full text of the GNU General Public License can be found in the doc 
 * directory.
 *
 * Copyright (C) 2002, 2003 Stefan Hoefer
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact Stefan Hoefer via email at stefan@hoefer.ch
 *
 ******************************************************************************/

#include "record.h"
#include "lineiterator.h"

CRecord::CRecord()
{
    Initialize();
}

void CRecord::Initialize()
{
    m_nId = 0;
    m_nSecondId = 0;
    m_lPosition = 0;
    m_nSize = 0;
}

void CRecord::Load(FILE *deskriptor)
{
    CStdioFile file(deskriptor);

    bool bReadOn = true;
    CStrings strLine;
    m_lPosition = ftell(deskriptor);
    m_nSize = 0;

    EmptyFields();

    do
    {
	if (file.ReadLine(strLine))
	{
	    bReadOn = ScanLine(strLine, true);
	}
	else
	{
	    bReadOn = false;
	}
    }while (bReadOn);
}

void CRecord::SetData(const char *strData)
{
    CStrings strRecord = strData;
    CLineIterator LineIt;

    LineIt << &strRecord;

    bool bReadOn = true;
    CStrings strLine;

    EmptyFields();

    do
    {
	strLine = LineIt++ + "\n";

	bReadOn = ScanLine(strLine, false);
    }while (bReadOn);
}

bool CRecord::ScanLine(CStrings &strLine, bool bSetSize)
{
    DecodeString(strLine);

    bool bReadOn = true;

    if (strLine[0] == '{') // Start of Record
    {
	int nPosition = strLine.Find(',');
	m_nId = atoi((const char *)strLine.Mid(1));

	if (nPosition > 0)
	{
	    m_nSecondId = atoi((const char *)strLine.Mid(nPosition+1));
	}
	else
	{
	    m_nSecondId = 0;
	}
	
	ASSERT(m_nId != 0);
    }
    else if (strLine[0] == '}') // End of Record
    {
	bReadOn = false;
    }
    else if (strLine[0] == '\n') // Empty Bytes
    {
	// Do nothing. These bytes are available for future write sessions.
    }
    else if (strLine[0] == '#') // Empty slot
    {
	// This should happen only on the first line.

	bReadOn = false;
    }
    else
    {
	CStrings strData;

	if (strLine.Length()-2 >= 4)
	{
	    strData = strLine.Mid(4,strLine.Length()-2); // This cuts the trailing '\n'
	}

	int nNumber = atoi((const char *)strLine.Left(4));
	
	CField *pField = Field(nNumber);
	
	if (pField == NULL)
	{
	    CField NewField;
	    
	    NewField.SetId(nNumber);
	    
	    pField = m_arrFields.Add(NewField);
	}
	
	pField->SetData(strData);
    }

    if (bSetSize)
    {
	m_nSize += strLine.Length();
    }

    return bReadOn;
}

void CRecord::EmptyFields()
{
    m_arrFields.reset();
}

CField *CRecord::Field(int nId) const
{
    CField Field;
    Field.SetId(nId);

    return m_arrFields.in(Field);
}

void CRecord::Set(int nId, const char *strData)
{
    CField *pField = Field(nId);

    if (pField == NULL)
    {
	CField Field;
	Field.SetId(nId);

	pField = m_arrFields.Add(Field);
    }

    pField->SetData(strData);
}

bool CRecord::Get(int nId, CStrings &strData) const
{
    CField *pField = Field(nId);

    if (pField)
    {
	strData = pField->Data();

	return true;
    }
    else
    {
	return false;
    }
}

int CRecord::ActualSize()
{
    return Data().Length();
}

CStrings CRecord::Data(int nSize)
{
    CStrings strData;

    strData.Format("{%d,%d\n", Id(), SecondId());
    
    for (int nIndex = 0; nIndex < m_arrFields.card(); nIndex++)
    {
	CField *pField = m_arrFields[nIndex];
	CStrings strLine;
	strLine.Format("%04d%s", pField->Id(), (const char *)pField->Data());

	EncodeString(strLine);

	strLine += "\n";

	strData += strLine;
    }

    if (strData.Length() < nSize-2)
    {
	int nPad = nSize-strData.Length()-2;

	for (int nIndex=0; nIndex < nPad; nIndex++)
	{
	    strData += '\n';
	}
    }

    strData += "}\n";

    return strData;
}

void CRecord::EncodeString(CStrings &strData)
{
    strData.Replace("\\", "\\\\");
    strData.Replace("\n", "\\n");
}

void CRecord::DecodeString(CStrings &strData)
{
    //strData.Replace("\\n", "\n");

    for (int nIndex = 0; nIndex < strData.Length(); nIndex++)
    {
	if (strData[nIndex] == '\\')
	{
	    if (nIndex + 1 < strData.Length())
	    {
		if (strData[nIndex+1] == '\\')
		{
		    strData = strData.Left(nIndex+1)+strData.Mid(nIndex+2);
		}
		else if (strData[nIndex+1] == 'n')
		{
		    strData = strData.Left(nIndex)+"\n"+strData.Mid(nIndex+2);
		}
	    }
	    
	}
    }
}

bool CRecord::operator==(const CRecord &record) const
{
    //STHTODO

    return false;
}










