/*******************************************************************************
 *
 * libbasic Version 0.0.1
 *
 * libbasic is a simple framework of foundation classes written in C++.
 *
 * The full text of the GNU General Public License can be found in the doc 
 * directory.
 *
 * Copyright (C) 2002, 2003 Stefan Hoefer
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * You can contact Stefan Hoefer via email at stefan@hoefer.ch
 *
 ******************************************************************************/

#ifndef UNIQUE_H

#define UNIQUE_H

//#include "container2.h"
#include <assert.h>
#include "identificator.h"
#include "basiclist.h"
#include "iterator.h"

class unique_object
{
 private:
/*    static long int next_number=0L; */
/*    static basic_list<long int> *lost_numbers=0; */
/*    static basic_list<long int> *found_numbers=0; */
    static /*long int*/ IDENT next_number;
    static basic_list</*long int*/IDENT> *lost_numbers;
    static basic_list</*long int*/IDENT> *found_numbers;
    /*long int*/ IDENT myid;
 public:
    unique_object();
    unique_object(/*long int*/IDENT);
    inline /*long int*/IDENT get_id() { return myid;}
    ~unique_object();
};

template <class node>
class unique_list;

// Caution: at this time, a unique_list_member may only be member of **ONE** unique_list at once (due to the lock flag). 

class unique_list_member
{
    friend class unique_list_dummy;
 private:
    /*long int*/IDENT m_Id;
    unsigned short locked;
 protected:
    virtual void NotifyIdChanges(IDENT id);
 public:
    inline /*long int*/IDENT get_id() const { return m_Id;}
    inline IDENT Id() const { return get_id(); }
    inline unsigned short is_locked() const { return locked;}
    unique_list_member();
    unique_list_member(/*long int*/IDENT id);
    void SetId(IDENT id);
};
 
class unique_list_dummy
{
 protected:
    void set_id(unique_list_member *, /*long int*/IDENT);
    void lock(unique_list_member *);
    void unlock(unique_list_member *);
};

template <class node>
class unique_list : public basic_list<node>, unique_list_dummy
{
 private:
    long int next_number;
    basic_list<IDENT> lost_numbers;
    basic_list<IDENT> found_numbers;
 private:
    bool CheckId(IDENT id);
 public:
    unique_list();
    ~unique_list();
    void operator+=(node n);
    void operator-=(node n);
    void reset();
    void delete_all();
    void get_members(basic_list<node> *l);
    void Merge( unique_list<node> *pList);
    node Find(IDENT id);
};

//-----------------------------------------------------------------------------------------------------
// Implementation Area
//-----------------------------------------------------------------------------------------------------

template <class node>
void unique_list<node>::Merge( unique_list<node> *pList)
{
    // for all element in the old list
	
    while ( node pNode = (*pList)[0] )
    {
	*pList -= pNode;	

	*this += pNode;		
    }	
}

template <class node>
node unique_list<node>::Find(IDENT id)
{
    iterator<node> NodeIt;

    NodeIt << this;

    while (node pNode = NodeIt++)
    {
	if (pNode->get_id() == id)
	{
	    return pNode;
	}
    }

    return NULL;
}

template <class node>
unique_list<node>::unique_list()
{
    next_number = 1L;
    //lost_numbers = new basic_list<IDENT>;
    //found_numbers = new basic_list<IDENT>;
}

template <class node>
unique_list<node>::~unique_list()
{
    //delete lost_numbers;
    //delete found_numbers;
}

template <class node>
bool unique_list<node>::CheckId(IDENT id)
{
    return !( (id==0)
	      || lost_numbers.in(id)
	      || (id<next_number && !found_numbers.in(id))
	      || (id==UNDEFINED_IDENT));
}

template <class node>
void unique_list<node>::operator+=(node n)
{ 
    assert(!n->is_locked()); // At this time, only this assertion is made. No further handling of this situation yet.
    
    lock(n);
    basic_list<node>::operator+=(n);

    if(!CheckId(n->get_id()))
    {  
	// We have to give it a new id
	if(found_numbers.card())
	{
	    set_id(n, found_numbers.first());
	    found_numbers -= n->get_id();
	    if(n->get_id()>=next_number)
	    {
		lost_numbers += n->get_id();
	    }	
	}
	else
	{
	    while(lost_numbers.in(next_number))
	    {
		lost_numbers -= next_number;
		next_number++;
	    }
	    if(found_numbers.in(next_number))
	    {
		found_numbers -= next_number;
	    }

	    set_id(n,next_number++);
	}
    }
    else
    {  
	// We can keep this id

	if(n->get_id()==next_number)
	{
	    next_number++;
	    if(found_numbers.in(n->get_id()))
	    {
		found_numbers -= n->get_id();
	    }
	}
	else
	{
	    lost_numbers += n->get_id();
	}
    }
}

template <class node>
void unique_list<node>::operator-=(node n)
{
    if(basic_list<node>::in(n))
    {          
	// makes this safe; otherwise, unlocks may be cheated!
	unlock(n);

	basic_list<node>::operator-=(n);
     
	if(lost_numbers.in(n->get_id()))
	{
	    lost_numbers -= n->get_id();
	}
	found_numbers += n->get_id();
    }
}

template <class node>
void unique_list<node>::reset()
{
    iterator<node> node_it;
    node n;
    node_it << this;
    while(n = node_it++)
    {
	unlock(n);
    }

    basic_list<node>::reset();

    lost_numbers.reset();
    found_numbers.reset();
    next_number = 1L;
}

template <class node>
void unique_list<node>::delete_all()
{
    iterator<node> node_it;
    node n;

    node_it << this;
    while(n = node_it++)
    {
	delete n;
    }

    basic_list<node>::reset();

    lost_numbers.reset();
    found_numbers.reset();
    next_number = 1L;
}

template <class node>
void unique_list<node>::get_members(basic_list<node> *l)     // not supported!
{
    ASSERT(false);
}

#endif



















